# -*- coding: utf-8 -*-
"""Utility functions for quantization hadamard transformation."""

import typing as tp

import scipy.linalg
import torch

from .functional import is_pow2

# Adapted from https://github.com/Cornell-RelaxML/quip-sharp/blob/main/lib/utils/matmul_had.py
# Copied from https://github.com/spcl/QuaRot/blob/main/fake_quant/hadamard_utils.py

__all__ = ["random_hadamard_matrix", "HadamardMatrix", "hardmard_transform"]


def _matmul_hadU(X: torch.Tensor, hadamard_K: torch.Tensor | None, K: int) -> torch.Tensor:
    """Apply Hadamard matrix to the input tensor."""
    n = X.shape[-1]
    input = X.clone().view(-1, n, 1)
    output = input.clone()
    while input.shape[1] > K:
        input = input.view(input.shape[0], input.shape[1] // 2, 2, input.shape[2])
        output = output.view(input.shape)
        output[:, :, 0, :] = input[:, :, 0, :] + input[:, :, 1, :]
        output[:, :, 1, :] = input[:, :, 0, :] - input[:, :, 1, :]
        output = output.view(input.shape[0], input.shape[1], -1)
        (input, output) = (output, input)
    del output

    if K > 1:
        assert hadamard_K is not None
        input = hadamard_K.view(1, K, K).to(input) @ input

    return input.view(X.shape) / torch.tensor(n).sqrt()


def random_hadamard_matrix(size: int) -> torch.Tensor:
    """Generate a random Hadamard matrix of size `size`."""
    # See https://cornell-relaxml.github.io/quip-sharp/ , Section "Randomized Hadamard Transformation"
    Q = torch.randint(low=0, high=2, size=(size,)).to(torch.float64)
    Q = Q * 2 - 1
    Q = torch.diag(Q)
    hadamard_K, K = HadamardMatrix.get_lhs(size)
    return _matmul_hadU(Q, hadamard_K=hadamard_K, K=K)


def hardmard_transform(
    x: torch.Tensor, hadamard_rhs: torch.Tensor, hadamard_lhs: torch.Tensor | None, lhs_k: int, scaled: bool = False
) -> torch.Tensor:
    """Apply Hadamard matrix to the input tensor."""
    shape = x.shape
    n = shape[-1]
    rhs_k = n // lhs_k
    x = x.view(-1, lhs_k, rhs_k)
    x = x @ hadamard_rhs  # hadamard_rhs is symmetric since rhs_k is a power of 2
    if not scaled:
        x = x.mul_((1.0 / torch.tensor(n, dtype=torch.float64).sqrt()).to(device=x.device, dtype=x.dtype))
    if lhs_k > 1:
        x = hadamard_lhs @ x
    return x.view(shape)


# region hadamard matrices
class HadamardMatrix:
    # hadamard matrices for had12, had36.pal2, had52,will, had60.pal, had108.pal, had140.pal, had156.will, had172.will:
    # http://www.neilsloane.com/hadamard/index.html

    _cache: tp.ClassVar[dict[tuple[int, bool, torch.dtype, torch.device], tuple[torch.Tensor, torch.Tensor, int]]] = {}

    @staticmethod
    def get(
        n: int, *, scale: bool, dtype: torch.dtype = torch.float32, device: torch.device | str = "cpu"
    ) -> tuple[torch.Tensor, torch.Tensor, int]:
        """Get the Hadamard matrix and left-hand-size dimension for a given input size."""
        device = torch.device(device)
        key = (n, scale, dtype, device)
        if key not in HadamardMatrix._cache:
            lhs, k = HadamardMatrix.get_lhs(n)
            rhs = torch.FloatTensor(scipy.linalg.hadamard(n // k))
            if scale:
                rhs = rhs.to(dtype=torch.float64).mul_(1.0 / torch.tensor(n, dtype=torch.float64).sqrt_())
            lhs = lhs.to(dtype=dtype, device=device)
            rhs = rhs.to(dtype=dtype, device=device)
            HadamardMatrix._cache[key] = rhs, lhs, k
        return HadamardMatrix._cache[key]

    @staticmethod
    def get_lhs(n: int) -> tuple[torch.FloatTensor, int]:
        """Get the Hadamard matrix (left-hand-side) and its dimension for a given input size."""
        for k in [172, 156, 140, 108, 60, 52, 36, 28, 40, 20, 12]:
            if n % k == 0:
                assert is_pow2(n // k)
                return HadamardMatrix._get_hadamard_k(k), k
        assert is_pow2(n)
        return torch.FloatTensor([1]), 1

    @staticmethod
    def _get_hadamard_k(k: int) -> torch.FloatTensor:
        if is_pow2(k):
            return torch.FloatTensor(scipy.linalg.hadamard(k))
        else:
            return getattr(HadamardMatrix, f"_get_hadamard_{k}")()

    @staticmethod
    def _get_hadamard_12() -> torch.FloatTensor:
        return torch.FloatTensor(
            [
                [+1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1],
                [+1, +1, -1, +1, -1, -1, -1, +1, +1, +1, -1, +1],
                [+1, +1, +1, -1, +1, -1, -1, -1, +1, +1, +1, -1],
                [+1, -1, +1, +1, -1, +1, -1, -1, -1, +1, +1, +1],
                [+1, +1, -1, +1, +1, -1, +1, -1, -1, -1, +1, +1],
                [+1, +1, +1, -1, +1, +1, -1, +1, -1, -1, -1, +1],
                [+1, +1, +1, +1, -1, +1, +1, -1, +1, -1, -1, -1],
                [+1, -1, +1, +1, +1, -1, +1, +1, -1, +1, -1, -1],
                [+1, -1, -1, +1, +1, +1, -1, +1, +1, -1, +1, -1],
                [+1, -1, -1, -1, +1, +1, +1, -1, +1, +1, -1, +1],
                [+1, +1, -1, -1, -1, +1, +1, +1, -1, +1, +1, -1],
                [+1, -1, +1, -1, -1, -1, +1, +1, +1, -1, +1, +1],
            ]
        )

    @staticmethod
    def _get_hadamard_40() -> torch.FloatTensor:
        return torch.FloatTensor(
            [
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
            ]
        )

    @staticmethod
    def _get_hadamard_20() -> torch.FloatTensor:
        return torch.FloatTensor(
            [
                [+1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1],
                [+1, +1, -1, +1, +1, -1, -1, -1, -1, +1, -1, +1, -1, +1, +1, +1, +1, -1, -1, +1],
                [+1, +1, +1, -1, +1, +1, -1, -1, -1, -1, +1, -1, +1, -1, +1, +1, +1, +1, -1, -1],
                [+1, -1, +1, +1, -1, +1, +1, -1, -1, -1, -1, +1, -1, +1, -1, +1, +1, +1, +1, -1],
                [+1, -1, -1, +1, +1, -1, +1, +1, -1, -1, -1, -1, +1, -1, +1, -1, +1, +1, +1, +1],
                [+1, +1, -1, -1, +1, +1, -1, +1, +1, -1, -1, -1, -1, +1, -1, +1, -1, +1, +1, +1],
                [+1, +1, +1, -1, -1, +1, +1, -1, +1, +1, -1, -1, -1, -1, +1, -1, +1, -1, +1, +1],
                [+1, +1, +1, +1, -1, -1, +1, +1, -1, +1, +1, -1, -1, -1, -1, +1, -1, +1, -1, +1],
                [+1, +1, +1, +1, +1, -1, -1, +1, +1, -1, +1, +1, -1, -1, -1, -1, +1, -1, +1, -1],
                [+1, -1, +1, +1, +1, +1, -1, -1, +1, +1, -1, +1, +1, -1, -1, -1, -1, +1, -1, +1],
                [+1, +1, -1, +1, +1, +1, +1, -1, -1, +1, +1, -1, +1, +1, -1, -1, -1, -1, +1, -1],
                [+1, -1, +1, -1, +1, +1, +1, +1, -1, -1, +1, +1, -1, +1, +1, -1, -1, -1, -1, +1],
                [+1, +1, -1, +1, -1, +1, +1, +1, +1, -1, -1, +1, +1, -1, +1, +1, -1, -1, -1, -1],
                [+1, -1, +1, -1, +1, -1, +1, +1, +1, +1, -1, -1, +1, +1, -1, +1, +1, -1, -1, -1],
                [+1, -1, -1, +1, -1, +1, -1, +1, +1, +1, +1, -1, -1, +1, +1, -1, +1, +1, -1, -1],
                [+1, -1, -1, -1, +1, -1, +1, -1, +1, +1, +1, +1, -1, -1, +1, +1, -1, +1, +1, -1],
                [+1, -1, -1, -1, -1, +1, -1, +1, -1, +1, +1, +1, +1, -1, -1, +1, +1, -1, +1, +1],
                [+1, +1, -1, -1, -1, -1, +1, -1, +1, -1, +1, +1, +1, +1, -1, -1, +1, +1, -1, +1],
                [+1, +1, +1, -1, -1, -1, -1, +1, -1, +1, -1, +1, +1, +1, +1, -1, -1, +1, +1, -1],
                [+1, -1, +1, +1, -1, -1, -1, -1, +1, -1, +1, -1, +1, +1, +1, +1, -1, -1, +1, +1],
            ]
        )

    @staticmethod
    def _get_hadamard_28() -> torch.FloatTensor:
        return torch.FloatTensor(
            [
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
            ]
        )

    @staticmethod
    def _get_hadamard_36() -> torch.FloatTensor:
        return torch.FloatTensor(
            [
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
            ]
        )

    @staticmethod
    def _get_hadamard_60() -> torch.FloatTensor:
        return torch.FloatTensor(
            [
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
            ]
        )

    @staticmethod
    def _get_hadamard_52() -> torch.FloatTensor:
        return torch.FloatTensor(
            [
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
            ]
        )

    @staticmethod
    def _get_hadamard_108() -> torch.FloatTensor:
        return torch.FloatTensor(
            [
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
            ]
        )

    @staticmethod
    def _get_hadamard_140() -> torch.FloatTensor:
        return torch.FloatTensor(
            [
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
            ]
        )

    @staticmethod
    def _get_hadamard_156() -> torch.FloatTensor:
        return torch.FloatTensor(
            [
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
            ]
        )

    @staticmethod
    def _get_hadamard_172() -> torch.FloatTensor:
        return torch.FloatTensor(
            [
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
                [
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    -1,
                ],
                [
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                ],
                [
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    -1,
                    +1,
                    -1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    -1,
                    +1,
                    -1,
                    +1,
                    +1,
                    +1,
                    +1,
                    -1,
                    -1,
                    +1,
                    +1,
                    -1,
                    -1,
                    -1,
                    +1,
                ],
            ]
        )


# endregion
# endregion
# endregion
